//     ======================= AssociativeMemory.h ==========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 03.02.2010
//     **********************************************************************

#if !defined(_AssociativeMemory_H_INCLUDED_v0_4_)
#define _AssociativeMemory_H_INCLUDED_v0_4_

#include "cv.h"
#include "cxmisc.h"
#include "..\Ver0.5\AssociativeMemory.h"

// Definition of search mode for associative memory
typedef enum {
	cLinearSearching = 0,
	cAdvancedSearching,
	cAdvancedSearchingTestMode,
	cQuickSearching,
	cQuickSearchingTestMode
} CvAM_SearchMode;

// Associative video memory consist of associative bases tree and use input
// image as a key for data access.


// Associative memory class for integer 32-bit data
class CvAssociativeMemory32S_v0_4 {
	void* p;
public:

	CvAssociativeMemory32S_v0_4(int aMemStorageBlockSize = 0);
	~CvAssociativeMemory32S_v0_4();
	
	// Creating of associative memory
	//
	// Notes: The parameter aKeyImgSize is depended on sequence 40, 80, 160, 320, 640, 1280... 2^n*10.
	//        Use of this numbers for sizing gives more accuracy in recognition.
	//        The parameter aLevelMax set a maximal level for associative tree.
	//        If set to zero then a maximal level will be computed as optimal for image key size.
	void Create(CvSize aKeyImgSize, short aLevelMax = 0);


	// Set an image for processing
	void SetImage(const void* apSrcImg, const void* apObjMaskImg = NULL);

	// Reading from associative memory cell (associative base)
	//
	// Notes: aInterestArea - interest area for object searching;
	//        apObjRect     - rectangle where object found;
	//        appData       - pointer to the data of associative cell;
	//        apIndex       - unique index of associative base;
	//        apHitCounter  - number of hitting to associative base;
	//        apSimilarity  - similarity of the interest area to an object (0 ... 1);
	//        aTotalSearch  - flag of the total search until end level of associative memory.
	bool Read(CvRect aInterestArea, CvRect* apObjRect, long** appData,
		      uint64* apIndex = NULL, uint64* apHitCounter = NULL,
			  double* apSimilarity = NULL, bool aTotalSearch = false);

	// Writing to associative memory cell
	void Write(CvRect aInterestArea, long* apData);

	// Object recognition
	void ObjectRecognition(CvRcgAMFunc32S RcgAMFunc, void *args = NULL);
	void ObjectTracking(CvRcgAMFunc32S RcgAMFunc, void *args = NULL,
		                bool   aTrainingDuringTracking = true,
		                double aSimilarityUpTrd = cAM_SimilarityUpTrd,
						double aSimilarityDnTrd = cAM_SimilarityDnTrd);

	// Optimization of associative tree
	void OptimizeAssociativeTree();
	
	// Get size of packed recognition data
	long GetPackedDataSize();
	// Writing a packed recognition data to memory
	void WritePackedData(void* apMem);
	// Reading a packed recognition data from memory
	void ReadPackedData(void* apMem);
	
	// Saving of recognition data
	bool Save(char* aFileName, bool aSaveWithoutOptimization = false);
	// Loading of recognition data
	bool Load(char* aFileName);

	// The estimation of an opportunity for training
	double EstimateOpportunityForTraining(CvRect aInterestArea) { return 0; }

	// Perform "AMFunc" for each associative base
	bool DoForEach(CvAMFunc32S AMFunc, void *args);
	
	// Get current index of associative base
	uint64 GetCurIndex();
	// Get total number of associative bases
	long   GetTotalABases();
	// Get total number of memory levels
	short  GetTotalLevels();
	// Get read-write counter
	uint64 GetWrRdCounter();
	// Get size of key image of associative memory
	CvSize GetKeyImageSize();
	// Set the search mode
	void SetSearchMode(CvAM_SearchMode aSearchMode);
	// Get search fail rate
	double GetSearchFailRate();
};

#endif // !defined(_AssociativeMemory_H_INCLUDED_v0_4_)
