//     ========================= MxValidate.h =============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 23.04.2009
//     **********************************************************************

#include "cv.h"
#include "highgui.h"
#include "cxmisc.h"
#include <stdio.h>
#include "..\..\..\include\Ver0.4\AssociativeMemory.h"


#define cRandMax     400
#define cClustersMax 300
#define cPointMax    50
#define cClsRadius   50
#define cSrcImgSize  (cRandMax*2)
#define cAVMKeySize  80

// Data transmission structure for "doSearchObject" function
struct TSearchObjectData {
	IplImage* pImg;
	bool      Hit;
};

void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args) 
{
	TSearchObjectData* d = (TSearchObjectData*)args;

	switch (aState) { 
		case cRecognizedObject:
			cvRectangle( d->pImg, cvPoint(apParam->ObjRect.x, apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,128,0),1,8,0);
			
			printf("Recognized, Id = %lu, Similarity = %f, rectFound(%d, %d, %d, %d)\n", 
				*apData, apParam->Similarity, 
				apParam->ObjRect.x, apParam->ObjRect.y, 
				apParam->ObjRect.width, apParam->ObjRect.height); 
			d->Hit = true;
			break; 

		case cGeneralizedObject: 
		case cTrackedObject:
			cvRectangle( d->pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,0,0),2,8,0);

			printf("Tracked/Generalized, Id = %lu, Similarity = %f, rectFound(%d, %d, %d, %d)\n", 
				*apData, apParam->Similarity, 
				apParam->ObjRect.x, apParam->ObjRect.y, 
				apParam->ObjRect.width, apParam->ObjRect.height); 
			d->Hit = true;
		  break; 
	} 
}


CvPoint GeneratePoint(CvPoint aCentral, int aDisp)
{
	CvPoint Point;
	Point.x = aCentral.x + (rand()%(aDisp+1)) - (aDisp>>1);
	Point.y = aCentral.y + (rand()%(aDisp+1)) - (aDisp>>1);

	return Point;
}

template <class AssociativeMemory32S>
void HitTest(AssociativeMemory32S*, int aVerNum, int aSampleMaxNum, int* apHitCount)
{
	int i,j;
	IplImage *SrcImg = 0, *SrcGrayImg = 0, *DstImg = 0;
	AssociativeMemory32S* am = new AssociativeMemory32S;
	
	cvNamedWindow("Points", 1);

	am->Create(cvSize(cAVMKeySize, cAVMKeySize));
	
	SrcImg = cvCreateImage(cvSize(cSrcImgSize, cSrcImgSize), IPL_DEPTH_8U, 3);
	SrcGrayImg = cvCreateImage(cvSize(cSrcImgSize, cSrcImgSize), IPL_DEPTH_8U, 1);
	DstImg = cvCreateImage(cvSize(320, 200), IPL_DEPTH_8U, 1);

	uint64 ABIndex;
	uint64 ABHitCounter;
	double Similarity;
	CvPoint ClusterPnt[cClustersMax];

	CvRect InterestArea;
	InterestArea.x = 40;
	InterestArea.y = 40;
	InterestArea.width  = 80;
	InterestArea.height = 80;
	long ObjIndex = 1;

	for(int s=0; s<aSampleMaxNum; s++) {
		cvSet(SrcImg, cvScalarAll(0),0);

		int Clusters = (cClustersMax/2) + (rand() % (cClustersMax/2));

		for(i=0; i<=Clusters; i++) {
			ClusterPnt[i] = GeneratePoint(cvPoint(cRandMax, cRandMax), cSrcImgSize);
		}
		
		for(i=0; i<=Clusters; i++) {
			for(j=0; j<cPointMax; j++) {
				cvCircle( SrcImg, GeneratePoint(ClusterPnt[i], cClsRadius), 1+(rand() % 33),
									  CV_RGB(rand()%0xFF,rand()%0xFF,rand()%0xFF),1, 8, 0 );
			}
		}
		cvSmooth(SrcImg, SrcImg,CV_GAUSSIAN, 9, 9, 9);
		cvCvtColor(SrcImg, SrcGrayImg, CV_BGR2GRAY);
		cvEqualizeHist(SrcGrayImg, SrcGrayImg);
		cvResize(SrcGrayImg, DstImg, CV_INTER_LINEAR );

		am->SetImage(DstImg);

		printf("{ %-9u }--------------------------------------------------------[v0.%u]--\n", s, aVerNum);

		// Training
		CvRect SAr;

		am->Write(InterestArea, &ObjIndex);

		if(am->Read(InterestArea, &SAr, 0, &ABIndex, &ABHitCounter, &Similarity, true)) {
			printf("Hitting! Id = %lu, Idx = %I64u, Hit = %I64u, Total = %lu, Similarity = %f\n",
				   ObjIndex, ABIndex, ABHitCounter, am->GetTotalABases(), Similarity);
					
		} else {
			printf("Fail: Id = %lu, Idx = %I64u, Hit = %I64u, Total = %lu, Similarity = %f\n",
				   ObjIndex, ABIndex, ABHitCounter, am->GetTotalABases(), Similarity);
		}

		ObjIndex++;

		TSearchObjectData d;
		d.pImg = DstImg;
		d.Hit  = false;

		am->ObjectRecognition(doSearchObject, &d);

		if(d.Hit) { (*apHitCount)++; }

		cvShowImage("Points",DstImg);

		if( cvWaitKey(10) >= 0 ) break;
	}

	delete am;
}

int main(int argc, char* argv[])
{
	int SampleMaxNum;
	int TestTotal = 100;
		
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Help for \"Matrix validate\" test ===\n\n"
			   "MxValidate.exe -Smp(<Number>)\n\n"
			   "   -Smp(<Number>) <Number> - total number of sample;\n\n"
			   "Example:\n\n"
			   "MxValidate.exe -Smp(200)\n\n");
		return 0;
	}	

	#define cStrMax 256
	char *pDataStr;
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(int i=1; i<argc; i++) strcat(CommandsStr,argv[i]);

	if(pDataStr = strstr(CommandsStr,"-Smp(")) {
				pDataStr += 5; sscanf(pDataStr,"%d",&SampleMaxNum);
				if(SampleMaxNum < TestTotal) {
					printf("Total number of sample is too small. Minimal is %d\n", TestTotal);
					return -1;
				}
	} else {
		printf("The key \"-Smp\" not found.\n");
		return -1;
	}

	int HitCount_v0_4 = 0;
	printf("\n================================= Ver 0.4 ===================================\n");
	HitTest((CvAssociativeMemory32S_v0_4*)(NULL), 4, SampleMaxNum, &HitCount_v0_4);

	int HitCount_v0_5 = 0;
	printf("\n================================= Ver 0.5 ===================================\n");
	HitTest((CvAssociativeMemory32S*)(NULL), 5, SampleMaxNum, &HitCount_v0_5);

	printf("\n\n Statistics of hits:\n\n");
	printf(" --- AVM v0.4 ... %u/%u [%5.1f %%]\n\n",
		HitCount_v0_4, SampleMaxNum, (double(HitCount_v0_4)/double(SampleMaxNum))*100.0);

	printf(" --- AVM v0.5 ... %u/%u [%5.1f %%]\n\n",
		HitCount_v0_5, SampleMaxNum, (double(HitCount_v0_5)/double(SampleMaxNum))*100.0);

	cvWaitKey();

	return 0;
}
