//     ========================= OpticalFlow.cpp ============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update: 30.05.2009
//     **********************************************************************

#pragma warning (disable:4996) // This function or variable may be unsafe

#include <stdio.h>
#include "highgui.h"
#include "..\..\..\include\Ver0.4\AssociativeMemory.h"
#include "..\..\..\include\OpticalFlow.h"

int main(int argc, char* argv[])
{
	int       i;
	long      FilmLengthMs  = 0;
	double    FilmFrameRate;
	IplImage  *OutImg, *gray, *frame;
	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Optical flow help ===\n\n"
			   "OpticalFlow.exe [file.avi | 0..9] [ /w | /c ]\n"
			   " 0..9 - capture device number\n"
			   " /w   - write result in file_out.avi\n"
			   " /c   - colored motion\n\n");
		return 0;
	}

	char* argv1 = _strlwr(argv[1]);
	if(!(strstr(argv1,".avi") || isdigit(argv[1][0]))){
		printf("AVI files are processing only!\n\n");
		return -1;
	}

	#define cStrMax 256
	char FileName[cStrMax];
	char OutStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(i=2; i<argc; i++) strcat(CommandsStr,argv[i]);
	memset(FileName,0,cStrMax);
	CvCapture*     InAVI;
	CvVideoWriter* OutAVI;

	bool WriteResultToFile = false;
	bool ColoredMotion     = false;

	if(strstr(CommandsStr,"/w"))  { WriteResultToFile = true; }
	if(strstr(CommandsStr,"/c"))  { ColoredMotion     = true; }

	int WIDTH, HEIGHT;
	CvSize ImgSize;

    if(isdigit(argv[1][0])) {
		printf("Processing capture...\n");
        InAVI = cvCaptureFromCAM( argc == 2 ? argv[1][0] - '0' : 0 );
		if( !InAVI ) {
			fprintf(stderr,"Could not initialize capturing...\n");
			return -1;
		}
		frame = cvQueryFrame(InAVI);
		ImgSize = cvGetSize(frame);
		WIDTH   = ImgSize.width;
		HEIGHT  = ImgSize.height;
		FilmFrameRate = 12;
    } else {
		printf("Processing avi file...\n");
		memcpy(FileName,argv1,strstr(argv1,".avi")-argv1);
		InAVI = cvCaptureFromFile(argv1);
		if(!InAVI) {
			printf("Can't find file %s.avi\n",FileName);
			return -1;
		}
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,1);
		FilmLengthMs  = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,0);
		WIDTH         = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_WIDTH));
		HEIGHT        = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_HEIGHT));
		FilmFrameRate = cvGetCaptureProperty(InAVI,CV_CAP_PROP_FPS);
		ImgSize       = cvSize(WIDTH, HEIGHT);
    }

	gray = cvCreateImage(ImgSize, 8, 1); 
	OutImg = cvCreateImage(ImgSize, 8, 3);

	if(WriteResultToFile) {
		sprintf(OutStr,"%s_out.avi",FileName);
		OutAVI = cvCreateVideoWriter(OutStr, CV_FOURCC('x','v','i','d'), FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		if(!OutAVI) {
			OutAVI = cvCreateVideoWriter(OutStr, -1, FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		}
	}

	CvOpticalFlow OptFlw(ImgSize);

	cvNamedWindow("Result", 1);

	for(;;) {
		
		frame = cvQueryFrame(InAVI);
		if( !frame ) break;

		cvFlip(frame);
		cvCopy(frame, OutImg);
		cvCvtColor(frame, gray, CV_BGR2GRAY);

		OptFlw.Calculate(gray);
		OptFlw.DrawFlow(OutImg, ColoredMotion);

		ULONG POS_MSEC = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		int hour        = int(POS_MSEC/(3600000));
		POS_MSEC        %= 3600000L;
		int minute      = int(POS_MSEC/(60000));
		POS_MSEC        %= 60000L;
		int second      = int(POS_MSEC/1000);
		int millisecond = int(POS_MSEC % 1000);
		printf("==> film current position: %2.2d:%2.2d:%2.2d.%3.3d\n",
				hour, minute, second, millisecond);

		
		cvShowImage("Result", OutImg);
		if(WriteResultToFile) { cvWriteFrame(OutAVI,OutImg); }

		if( cvWaitKey(10) >= 0 ) break;
	}
	
	if(WriteResultToFile) { cvReleaseVideoWriter(&OutAVI); }
	cvReleaseCapture(&InAVI);
	cvReleaseImage(&gray);
	cvReleaseImage(&OutImg);
	
	return 0;
}