//     ======================== ROC_test.cpp ================================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 10.11.2009
//     **********************************************************************

#include "..\..\include\Ver0.4\AssociativeMemory.h"
#include "..\..\include\AVM_drawing.h"

#define cObjectDirName     "grey"
#define cMaskDirName       "mask"
#define cROC_ImageFileName "05vs03vsSE.jpg"
#define cROC_WinName       "AVM ROC curves"
#define cPairsTotal        36
#define cDiagramStepTotal  1000
#define cFloatMistake      1e-12

#define cBorder_Left       71
#define cBorder_Right      27
#define cBorder_Top        100
#define cBorder_Bottom     52

#define cDiagramXSize      1182
#define cDiagramYSize      587

#define cDiagramXStep      197
#define cDiagramYStep      196

#define cExpCorrRate 5.5
#define cBalanceRate 0.57
#define cIntervalX   7
#define cIntervalY   4
#define cVerBase     4

#define cStrMax 256
#define cImgW   768
#define cImgH   576

static double TabIntervalX[cIntervalX] = { 0.000001, 0.00001, 0.0001, 0.001, 0.01, 0.1, 1.0 + cFloatMistake };
static double TabIntervalY[cIntervalY] = { 0.001, 0.01, 0.1, 1.0 + cFloatMistake };

inline double GetDiagramExpRate(double aPercent, double aDiagramExpDiv) {
	return ((1.0 - exp(-aPercent*cExpCorrRate))/aDiagramExpDiv)*cBalanceRate + aPercent*(1.0 - cBalanceRate);
}

CvPoint GetDiagramPoint(double aPercentX, double aPercentY) {
	CvPoint Point; int i;
	static double DiagramExpDiv = 0;

	if(!DiagramExpDiv) { DiagramExpDiv = 1.0 - exp(-cExpCorrRate); }

	Point.x = cBorder_Left;
	Point.y = cDiagramYSize + cBorder_Top;
		
	if(aPercentX < TabIntervalX[0]) {
		aPercentX = 0;	
	} else {
		for(i=1; i<cIntervalX; i++) {
			if(aPercentX < TabIntervalX[i] && fabs(aPercentX - TabIntervalX[i]) > cFloatMistake) {
				aPercentX = (aPercentX - TabIntervalX[i-1])/(TabIntervalX[i] - TabIntervalX[i-1]); break;
			} else { Point.x += cDiagramXStep; }
		}
	}

	if(aPercentY < TabIntervalY[0]) {
		aPercentY = 0;	
	} else {
		for(i=1; i<cIntervalY; i++) {
			if(aPercentY < TabIntervalY[i] && fabs(aPercentY - TabIntervalY[i]) > cFloatMistake) {
				aPercentY = (aPercentY - TabIntervalY[i-1])/(TabIntervalY[i] - TabIntervalY[i-1]); break;
			} else { Point.y -= cDiagramYStep; }
		}
	}

	Point.x += int(cDiagramXStep*GetDiagramExpRate(aPercentX, DiagramExpDiv));
	Point.y -= int(cDiagramYStep*GetDiagramExpRate(aPercentY, DiagramExpDiv));

	return Point;
}

CvRect GetRect(IplImage *mask, IplImage *img)
{

      if (mask == NULL)
      {
            return cvRect(0, 0, img->width, img->height);
      }

      unsigned char *data = (unsigned char *)mask->imageData;

      int tx = mask->width;
      int ty = mask->height;
      int bx = 0;
      int by = 0;

      bool isData = false;

      for(int i = 0; i < mask->height; i++)
      {
            for(int j = 0; j < mask->width; j++)

            {
                  if (data[i * mask->widthStep + j])
                  {
                        if (i < ty) ty = i;
                        if (j < tx) tx = j;
                        if (i > by) by = i;
                        if (j > bx) bx = j;
                        isData = true;
                  }
            }
      }

      if (isData == false)
      {
            return cvRect(0, 0, img->width, img->height);
      }

      CvRect rect = cvRect(tx, ty, bx - tx, by - ty);

      if (rect.width > rect.height)
      {
            int d = (rect.width - rect.height) / 2;
            rect.width = rect.height;
            rect.x += d;
      }
      else if (rect.width < rect.height)
      {
            int d = (rect.height - rect.width) / 2;
            rect.height = rect.width;
            rect.y += d;
      }

      return rect;
}

// Prepare input image for testing
void PrepareImage(char* aDirName, long aSizeDiv, IplImage* aImage,
				  int aObjNum, int aFileIdx, CvRect* apInterestArea = NULL,
				  bool* apMaskFlag = NULL, IplImage* aMaskImage = NULL)
{
char FileNameStr[cStrMax];
CvRect Rect; IplImage *InputImg, *MaskImg = NULL;

	if(aSizeDiv == 1) {
		sprintf(FileNameStr,"%s\\%s\\%u\\%u_r%u.png", aDirName,
					cObjectDirName, aObjNum, aObjNum, aFileIdx);
	} else {
		sprintf(FileNameStr,"%s\\%s%u\\%u\\%u_r%u.png", aDirName,
					cObjectDirName, aSizeDiv, aObjNum, aObjNum, aFileIdx);
	}

	if(apMaskFlag) { *apMaskFlag = false; }
	InputImg = cvLoadImage(FileNameStr, 0);

	if(InputImg) {

		if(aFileIdx) {
			if(aSizeDiv == 1) {
				sprintf(FileNameStr,"%s\\%s\\%u\\%u_r%u.png", aDirName,
						cMaskDirName, aObjNum, aObjNum, aFileIdx);
			} else {
				sprintf(FileNameStr,"%s\\%s%u\\%u\\%u_r%u.png", aDirName,
						cMaskDirName, aSizeDiv, aObjNum, aObjNum, aFileIdx);
			}
		} else {
			// Using of mask image for rotation of 0 degrees: "*_c1.png"
			if(aSizeDiv == 1) {
				sprintf(FileNameStr,"%s\\%s\\%u\\%u_c1.png", aDirName,
						cMaskDirName, aObjNum, aObjNum);
			} else {
				sprintf(FileNameStr,"%s\\%s%u\\%u\\%u_c1.png", aDirName,
						cMaskDirName, aSizeDiv, aObjNum, aObjNum);
			}
		}

		MaskImg = cvLoadImage(FileNameStr, 0);

		if(MaskImg) {
			if(apMaskFlag) { *apMaskFlag = true; }
		} else {
			// >>> Make a mask by image threshold <<<
			MaskImg = cvCloneImage(InputImg);
			cvThreshold(MaskImg, MaskImg, 27, 255, CV_THRESH_BINARY);
			cvThreshold(MaskImg, MaskImg, 25, 0, CV_THRESH_TOZERO);
		}

		if(apInterestArea) {
			Rect = GetRect(MaskImg, InputImg);
			// >>> Increase the rectangle on 12% <<<
			int Width  = int(double(Rect.width)*1.12);
			int Height = int(double(Rect.height)*1.12);
			int x1 = Rect.x - ((Width - Rect.width)>>1);
			int y1 = Rect.y - ((Height - Rect.height)>>1);
			int x2 = x1 + Width;
			int y2 = y1 + Height;
			
			if(x1 < 0) { x1=0; }
			if(y1 < 0) { y1=0; }
			if(x2 > InputImg->width)  { x2 = InputImg->width; }
			if(y2 > InputImg->height) { y2 = InputImg->height; }
			
			Rect = cvRect(x1, y1, x2-x1, y2-y1);
			
			if(Rect.width != Rect.height) {
				int D = MIN(Rect.width, Rect.height);
				Rect.width  = D;
				Rect.height = D;
			}
		}
		/*
		// >>> Set the gray background instead black <<<
		cvXorS(MaskImg, cvScalarAll(0xFF), MaskImg);
		cvThreshold(MaskImg, MaskImg, 1, 64, CV_THRESH_BINARY);
		cvCopy(MaskImg, InputImg, MaskImg);
		*/
		if(InputImg->width != cImgW || InputImg->height != cImgH) {
			if(apInterestArea) {
				double ScaleX = cImgW/double(InputImg->width);
				double ScaleY = cImgH/double(InputImg->height);
				Rect.x = int(Rect.x*ScaleX);
				Rect.y = int(Rect.y*ScaleY);
				Rect.width = int(Rect.width*ScaleX);
				Rect.height = int(Rect.height*ScaleY);
				*apInterestArea = Rect;
			}
			cvResize(InputImg, aImage, CV_INTER_LINEAR);
			if(aMaskImage) { cvResize(MaskImg, aMaskImage, CV_INTER_LINEAR); }
		} else {
			cvCopy(InputImg, aImage);
			if(apInterestArea) { *apInterestArea = Rect; }
		}

		cvReleaseImage(&MaskImg);
		cvReleaseImage(&InputImg);
	} else { 
		printf("Can't find file %s\n",FileNameStr);
		exit(-1);	
	}
}

// Data transmission structure for "doSearchObject" function
struct TSearchObjectData {
	IplImage* pImg;
	bool      ObjectFound;
	double    MaxSimilarity;
};

void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void *args)
{
      TSearchObjectData* d = (TSearchObjectData*)args;

      switch(aState) {
            case cRecognizedObject:
            case cGeneralizedObject:
            case cTrackedObject:

				cvRectangle( d->pImg, cvPoint(apParam->ObjRect.x, apParam->ObjRect.y),
									  cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
											  apParam->ObjRect.y + apParam->ObjRect.height),
											  cvScalarAll(0xFF),1,8,0);
				d->ObjectFound = true;

				printf("Object: %u:%u\n", (*apData)>>6, (*apData) & 0x3F);

			case cUnknownObject:

				if(d->MaxSimilarity < apParam->Similarity) {
					d->MaxSimilarity = apParam->Similarity;
				}
                break;
      }
}

// Template of AVM testing 
template <class AssociativeMemory32S>
void MakeTest(AssociativeMemory32S*, char* aDirStr,
			  long aObjTotalNumber, long aSizeDiv, bool aMaskMode,
			  bool aTestOfSaving, bool aExtraData, double aRcgMxCellBase, int aVerIdx)
{
	int i,j;
	double Rate;
	CvRect InterestArea;
	IplImage *gray, *mask;
	IplImage* ResImage = 0;
	char OutStr[cStrMax];

	int ResImgW = cDiagramXSize + cBorder_Left + cBorder_Right;
	int ResImgH = cDiagramYSize + cBorder_Top  + cBorder_Bottom;
	CvFont BigFont;
	cvInitFont( &BigFont, 2, 1.0, 1.0, 0, 2, 8 );
	CvFont MiddleFont;
	cvInitFont( &MiddleFont, 2, 0.8, 0.8, 0, 2, 8 );
	CvFont SmallFont;
	cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );

	gray = cvCreateImage(cvSize(cImgW, cImgH), IPL_DEPTH_8U, 1);
	mask = (aMaskMode)?cvCreateImage(cvSize(cImgW, cImgH), IPL_DEPTH_8U, 1):NULL;

	sprintf(OutStr,"AVM_v0-%u_ROC_curves_obj_%u.png", cVerBase, aObjTotalNumber);
	ResImage = cvLoadImage(OutStr);
	if(ResImage) {
		if(!aVerIdx) {
			cvNamedWindow(cROC_WinName, 1);
			cvShowImage(cROC_WinName, ResImage);
			cvReleaseImage(&ResImage);
			return;
		}
	} else {
		ResImage = cvCreateImage(cvSize(ResImgW, ResImgH), IPL_DEPTH_8U, 3);
		cvSet(ResImage,cvScalarAll(0xFF),0);

		IplImage* ROC_Image = cvLoadImage(cROC_ImageFileName);

		if(ROC_Image) {
			if(ResImgW < ROC_Image->width || ResImgH > ROC_Image->height) {
				printf("Incorrect image size of file %s\n", cROC_ImageFileName);
				exit(-1);
			}
			#define cSubTop 16
			IplImage* tImage = cvCreateImage(cvSize(ResImgW, ResImgH - cBorder_Top + cSubTop), IPL_DEPTH_8U, 3);
			cvSetImageROI(ROC_Image, cvRect(0, ROC_Image->height - tImage->height, ResImgW, tImage->height));
			cvCopy(ROC_Image, tImage);
			cvResetImageROI(ROC_Image);
			
			cvSetImageROI(ResImage, cvRect(0, ResImage->height - tImage->height, ResImgW, tImage->height));
			cvCopy(tImage, ResImage);
			cvResetImageROI(ResImage);
			cvReleaseImage(&tImage);
			cvReleaseImage(&ROC_Image);
			cvPutText(ResImage, "AVM ROC curves", cvPoint(550, 70), &BigFont, cvScalarAll(0));
			cvPutText(ResImage, "Red   curve - Ver. 0.5-r6;",  cvPoint(10, 20), &SmallFont, CV_RGB(0xCF,0,0));
			cvPutText(ResImage, "Green curve - Ver. 0.4;",     cvPoint(10, 40), &SmallFont, CV_RGB(0,0x9F,0)); 
		} else {
			printf("Can't find file %s\n", cROC_ImageFileName);
			exit(-1);
		}

		// Drawing of points on axes
		for(i=0; i<cIntervalX; i++) {
			for(Rate = 0, j=0; j<10; j++) {
				Rate += TabIntervalX[i];
				CvPoint Pnt = GetDiagramPoint(Rate, 0);
				cvCircle( ResImage, Pnt, 4, CV_RGB(0,0,0xCF), 1, 8, 0 );
			}
		}

		for(i=0; i<cIntervalY; i++) {
			for(Rate = 0, j=0; j<10; j++) {
				Rate += TabIntervalY[i];
				CvPoint Pnt = GetDiagramPoint(0, Rate);
				cvCircle( ResImage, Pnt, 4, CV_RGB(0,0,0xCF), 1, 8, 0 );
				if(!i) {
					sprintf(OutStr, "%3.1f", Rate*100.0);
					cvPutText(ResImage, OutStr, cvPoint(10, Pnt.y), &SmallFont, CV_RGB(0,0,0));
				}
			}
		}
	}

	int ObjNum;
	int GenuineErr     = 0;
	int ImpostorErr    = 0;
	int GenuineIdx     = 0;
	int ImpostorIdx    = 0;
	int GenuinesTotal  = aObjTotalNumber*cPairsTotal;
	int ImpostorsTotal = aObjTotalNumber*(aObjTotalNumber-1)*cPairsTotal;
	
	double* pGenuineSim = (double*)cvAlloc(GenuinesTotal*sizeof(double));
	double* pImpostorSim = (double*)cvAlloc(ImpostorsTotal*sizeof(double));

	AssociativeMemory32S* am = new AssociativeMemory32S;

	cvNamedWindow("Input image", 1);

	#ifdef Only_AVM_v0_5
	CvDrawAssociativeTree DrawTree(am);
	#endif

	printf("AVM testing is started...\n");

    // Learning and saving all objects
	for(ObjNum=1; ObjNum<=aObjTotalNumber; ObjNum++) {
		#define cFileStr "Obj_%u_v0.%u.avm"
		sprintf(OutStr, cFileStr, ObjNum, aVerIdx + cVerBase);
		FILE* f = fopen(OutStr,"rb");
		if(f) { fclose(f); continue; }

		am->Create(cvSize(80, 80));

		#ifdef Only_AVM_v0_5
		am->SetParam(ptRcgMxCellBase, aRcgMxCellBase);
		#endif

		// Learn object
		for(i=0; i<cPairsTotal; i++) {
			bool MaskFlag;
			PrepareImage(aDirStr, aSizeDiv, gray, ObjNum, i*10, &InterestArea, &MaskFlag, mask);
			
			am->SetImage(gray, mask);
			long Data = (ObjNum<<6) | (i+1);
			am->Write(InterestArea, &Data);

			sprintf(OutStr, "Learning of object: %u:%2.2u  Bases: %-3lu", ObjNum, i+1, am->GetTotalABases());
			cvPutText(gray, OutStr, cvPoint(10, 40), &BigFont, cvScalarAll(0xFF));
			if(!MaskFlag) {
				cvPutText(gray, "Warning: Object mask is not found!", cvPoint(10, 80), &BigFont, cvScalarAll(0xFF));
			}
			cvRectangle(gray, cvPoint(InterestArea.x,  InterestArea.y),
							  cvPoint(InterestArea.x + InterestArea.width,
									  InterestArea.y + InterestArea.height), cvScalarAll(0xFF),2,8,0);
			#ifdef Only_AVM_v0_5
			static short DivCounter = 0;
			if(!(DivCounter++ % 5)) {
				DrawTree.ShowAssociativeTree();
			}
			#endif

			sprintf(OutStr, "AVM v0.%u saving...", aVerIdx + cVerBase);
			cvPutText(gray, OutStr, cvPoint(10, cImgH - 20), &MiddleFont, cvScalarAll(0xFF));

			cvShowImage("Input image", gray);				
			cvWaitKey(40);
		}

		#ifdef Only_AVM_v0_5
		DrawTree.ShowAssociativeTree();
		#endif

		sprintf(OutStr, cFileStr, ObjNum, aVerIdx + cVerBase);
		am->Save(OutStr, true);

		// Test of saving
		if(aTestOfSaving) {
			TSearchObjectData d;
			double SimArrayBefore[cPairsTotal];
			double SimArrayAfter[cPairsTotal];

			for(i=0; i<cPairsTotal; i++) {
					d.pImg = gray;
					d.MaxSimilarity = 0;

					PrepareImage(aDirStr, aSizeDiv, gray, ObjNum, i*10 + 5);
					am->SetImage(gray);
					am->ObjectRecognition(doSearchObject, &d);


					sprintf(OutStr, "Before saving: %u:%2.2u  Bases: %-3lu Sim: %6.4f",
												ObjNum, i+1, am->GetTotalABases(), d.MaxSimilarity);

					SimArrayBefore[i] = d.MaxSimilarity;

					cvPutText(gray, OutStr, cvPoint(10, 40), &BigFont, cvScalarAll(0xFF));

					cvShowImage("Input image", gray);				
					cvWaitKey(1);
			}

			#ifdef Only_AVM_v0_5
			am->Destroy();
			#endif

			sprintf(OutStr, cFileStr, ObjNum, aVerIdx + cVerBase);
			am->Load(OutStr);

			for(i=0; i<cPairsTotal; i++) {
					d.pImg = gray;
					d.MaxSimilarity = 0;

					PrepareImage(aDirStr, aSizeDiv, gray, ObjNum, i*10 + 5);
					am->SetImage(gray);
					am->ObjectRecognition(doSearchObject, &d);


					sprintf(OutStr, "After saving: %u:%2.2u  Bases: %-3lu Sim: %6.4f",
												ObjNum, i+1, am->GetTotalABases(), d.MaxSimilarity);

					SimArrayAfter[i] = d.MaxSimilarity;

					cvPutText(gray, OutStr, cvPoint(10, 40), &BigFont, cvScalarAll(0xFF));

					cvShowImage("Input image", gray);				
					cvWaitKey(1);
			}

			double SimDiff1, SimDiff2, SimDiffTotal = 0;

			for(i=0; i<cPairsTotal; i += 2) {

				SimDiff1 = SimArrayBefore[i] - SimArrayAfter[i];
				sprintf(OutStr, "SimDiff[%2u] = %8.4f", i, SimDiff1);
				cvPutText(gray, OutStr,  cvPoint(10, 70 + i*10), &SmallFont, cvScalarAll(0xFF));

				SimDiff2 = SimArrayBefore[i+1] - SimArrayAfter[i+1];
				sprintf(OutStr, "SimDiff[%2u] = %8.4f", i+1, SimDiff2);
				cvPutText(gray, OutStr,  cvPoint(550, 70 + i*10), &SmallFont, cvScalarAll(0xFF));

				SimDiffTotal += fabs(SimDiff1) + fabs(SimDiff2);
			}

			sprintf(OutStr, "SimDiffTotal = %12.10f", SimDiffTotal);
			cvPutText(gray, OutStr,  cvPoint(10, cImgH - 60), &SmallFont, cvScalarAll(0xFF));

			cvPutText(gray, "Press any key to continue", cvPoint(10, cImgH - 20), &MiddleFont, cvScalarAll(0xFF));
			cvShowImage("Input image", gray);
			cvWaitKey();
		}
	}

	FILE *fl = NULL, *flt = NULL;

	if(aExtraData) {
		// Open extra files
		if(aVerIdx == 0) {
			fl = fopen("ver04.roc", "wb");
			flt = fopen("ver04.txt", "wt");
		} else	{
			fl = fopen("ver05r4.roc", "wb");
			flt = fopen("ver05r4.txt", "wt");
		}

		int totalN = aObjTotalNumber*cPairsTotal + aObjTotalNumber*(aObjTotalNumber - 1)*cPairsTotal;

		fwrite(&totalN, sizeof(int), 1, fl);
	}

	// Loop of testing
	for(ObjNum=1; ObjNum<=aObjTotalNumber; ObjNum++) {

		sprintf(OutStr, cFileStr, ObjNum, aVerIdx + cVerBase);
		if(!am->Load(OutStr)) { continue; }

		// Loop of object recognition
		for(long rObjNum=1; rObjNum<=aObjTotalNumber; rObjNum++) {
			for(i=0; i<cPairsTotal; i++) {
				TSearchObjectData d;
				d.pImg = gray;
				d.ObjectFound = false;
				d.MaxSimilarity = 0;

				PrepareImage(aDirStr, aSizeDiv, gray, rObjNum, i*10 + 5);
				am->SetImage(gray);
				am->ObjectRecognition(doSearchObject, &d);

				int sim = int(d.MaxSimilarity * 10000);
				int index1 = 0;
				int index2 = 0;
				int isSame = 1;

				if(rObjNum == ObjNum) {
					if(!d.ObjectFound) { GenuineErr++; }
					sprintf(OutStr, "Genuine: %u:%2.2u  Bases: %-3lu Sim: %8.6f",
						    rObjNum, i+1, am->GetTotalABases(), d.MaxSimilarity);

					pGenuineSim[GenuineIdx++] = d.MaxSimilarity;
				} else {
					if(d.ObjectFound) { ImpostorErr++; }
					sprintf(OutStr, "Impostor: %u:%2.2u  Bases: %-3lu Sim: %8.6f",
						    rObjNum, i+1, am->GetTotalABases(), d.MaxSimilarity);

					pImpostorSim[ImpostorIdx++] = d.MaxSimilarity;
					isSame = 0;
				}

				// Save data into extra files
				if(fl) {
					fwrite(&index1, sizeof(int), 1, fl);
					fwrite(&index2, sizeof(int), 1, fl);
					fwrite(&isSame, sizeof(int), 1, fl);
					fwrite(&sim,    sizeof(int), 1, fl);
				}

				if(flt) { fprintf(flt, "%d\t%d\t%f\n", isSame, sim, d.MaxSimilarity); }

				cvPutText(gray, OutStr, cvPoint(10, 40), &BigFont, cvScalarAll(0xFF));

				double Progress = double(GenuineIdx+ImpostorIdx)/double(GenuinesTotal+ImpostorsTotal);
				sprintf(OutStr, "AVM v0.%u progress: %3.0f%%  Object: %u", aVerIdx + cVerBase, Progress*100.0, ObjNum);
				cvPutText(gray, OutStr, cvPoint(10, cImgH - 20), &MiddleFont, cvScalarAll(0xFF));

				sprintf(OutStr, "Genuines error:  %5.1f %%", (double(GenuineErr)/double(GenuineIdx))*100.0);
				cvPutText(gray, OutStr,  cvPoint(10, 70), &SmallFont, cvScalarAll(0xFF));

				sprintf(OutStr, "Impostors error: %5.1f %%",
					    (!ImpostorIdx)?0:((double(ImpostorErr)/double(ImpostorIdx))*100.0));
				cvPutText(gray, OutStr,  cvPoint(10, 90), &SmallFont, cvScalarAll(0xFF));

				cvShowImage("Input image", gray);

				cvWaitKey(1);
			}
			
			#ifdef Only_AVM_v0_5
			DrawTree.ShowAssociativeTree();
			#endif
		}
	}

	if(fl)  { fclose(fl);  }
	if(flt) { fclose(flt); }

	delete am;
	
	// Printing of statistics
	#ifndef Only_AVM_v0_5
	if(aVerIdx) {
	#endif

		sprintf(OutStr, "%5.1f %%", (double(GenuineErr)/double(GenuinesTotal))*100.0);
		cvPutText(ResImage, OutStr,  cvPoint(1016, 20), &SmallFont, CV_RGB(0xCF,0,0));
		sprintf(OutStr, "%5.1f %%", (double(ImpostorErr)/double(ImpostorsTotal))*100.0);
		cvPutText(ResImage, OutStr,  cvPoint(1016, 40), &SmallFont, CV_RGB(0xCF,0,0));

	#ifndef Only_AVM_v0_5
	} else {
	#endif

		if(aMaskMode) { cvPutText(ResImage, "Mask mode on", cvPoint(870, 60), &SmallFont, CV_RGB(0xCF,0,0)); }
		sprintf(OutStr, "Genuines total:  %u   Objects total: %u", GenuinesTotal, aObjTotalNumber);
		cvPutText(ResImage, OutStr,  cvPoint(300, 20), &SmallFont, CV_RGB(0,0,0));
		sprintf(OutStr, "Impostors total: %u", ImpostorsTotal);
		cvPutText(ResImage, OutStr,  cvPoint(300, 40), &SmallFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "Genuines error:",  cvPoint(870, 20), &SmallFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "Impostors error:",  cvPoint(870, 40), &SmallFont, CV_RGB(0,0,0));

	#ifndef Only_AVM_v0_5
		sprintf(OutStr, "%5.1f %%", (double(GenuineErr)/double(GenuinesTotal))*100.0);
		cvPutText(ResImage, OutStr,  cvPoint(1090, 20), &SmallFont, CV_RGB(0,0x9F,0));
		sprintf(OutStr, "%5.1f %%", (double(ImpostorErr)/double(ImpostorsTotal))*100.0);
		cvPutText(ResImage, OutStr,  cvPoint(1090, 40), &SmallFont, CV_RGB(0,0x9F,0));

	}
	#endif

	// Drawing of ROC curves
	double SimMin = 1.0;
	double SimMax = 0.0;

	for(i=0; i<GenuinesTotal; i++) {
		if(SimMin > pGenuineSim[i]) { SimMin = pGenuineSim[i]; }
		if(SimMax < pGenuineSim[i]) { SimMax = pGenuineSim[i]; }
	}

	for(i=0; i<ImpostorsTotal; i++) {
		if(SimMin > pImpostorSim[i]) { SimMin = pImpostorSim[i]; }
		if(SimMax < pImpostorSim[i]) { SimMax = pImpostorSim[i]; }
	}

	double SimStep = (SimMax - SimMin)/cDiagramStepTotal;
	double SimTrh  = SimMin;
	CvPoint PrevPoint;

	cvNamedWindow(cROC_WinName, 1);

	for(i=0; i<cDiagramStepTotal; i++, SimTrh += SimStep) {
		double FP =0, FN = 0;

		for(j=0; j<ImpostorsTotal; j++) {
			if(SimTrh < pImpostorSim[j]) { FP += 1.0; }
		}

		for(j=0; j<GenuinesTotal; j++) {
			if(SimTrh >= pGenuineSim[j]) { FN += 1.0; }
		}

		double rX = FP/double(ImpostorsTotal);
		double rY = FN/double(GenuinesTotal);
		CvPoint Point = GetDiagramPoint(rX, rY);

		double cH = (SimTrh < 0.5)?0:0x9F;
		CvScalar Color = (aVerIdx)?(CV_RGB(0xFF,cH,cH)):(CV_RGB(0,0x9F,0));

		if(i) { cvLine(ResImage, PrevPoint, Point, Color, 1, CV_AA, 0); }
		PrevPoint = Point;
		cvCircle( ResImage, Point, 2, Color, CV_FILLED, 8, 0);
		cvShowImage(cROC_WinName, ResImage);
		cvWaitKey(10);
	}

	sprintf(OutStr,"AVM_v0-%u_ROC_curves_obj_%u.png",aVerIdx + cVerBase, aObjTotalNumber);
	if(cvSaveImage(OutStr, ResImage)) {
		printf("\nResult image was saved into file %s\n\n", OutStr);
	}

	if(aMaskMode) { cvReleaseImage(&mask); }

	cvReleaseImage(&gray);
	cvReleaseImage(&ResImage);
	cvFree(&pGenuineSim);
	cvFree(&pImpostorSim);
	
	if(aVerIdx) { cvWaitKey(); }
}

int main(int argc, char* argv[])
{
	long SizeDiv = 1;
	long ObjTotalNumber;
	bool MaskMode      = false;
	bool TestOfSaving  = false;
	bool ExtraData     = false;
	long RcgMxCellBase = 20000;
	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Help for \"Associative video memory\" ROC test ===\n\n"
			   "ROC_test.exe -DBPath(<Object data base directory>) -Obj(<Number>) -SizeDiv(<D>)\n\n"
			   " Please, download black-and-white object and mask images\n"
			   "  form data base at http://staff.science.uva.nl/~aloi/\n"
			   "  Unpack it to your HDD and then point path to \"%s\" and \"%s\" directories.\n"
			   "   -SizeDiv(<D>)  <D = 1> - Image size divider (grey, grey2, grey4 - 1,2,4);\n"
			   "   -Obj(<Number>) <Number> - Total number of objects for testing;\n"
			   "   -Mask  - Switch on mask mode;\n"
			   "   -SvTst - Test of saving;\n"
			   "   -SvExt - Save the ROC files with extra data;\n"
			   "   -RcgMxCellBase(<Value>) <Value = 20000> - Base cell value for\n"
			   "                         recognition threshold calculation (only for AVM v0.5).\n\n"
			   "Example:\n\n"
			   "ROC_test.exe -DBPath(E:\\Objects) -Obj(64)\n\n",cObjectDirName, cMaskDirName);
		return 0;
	}

	char *pDataStr, *pEndDataStr;
	char DirStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(DirStr,0,cStrMax);
	memset(CommandsStr,0,cStrMax);
	for(int i=1; i<argc; i++) strcat(CommandsStr,argv[i]);

	if(pDataStr = strstr(CommandsStr,"-DBPath(")) {
				pDataStr += 8; pEndDataStr = strstr(pDataStr,")");
				memcpy(DirStr, pDataStr, pEndDataStr-pDataStr);
	} else {
		printf("The key \"-DBPath\" not found.\n");
		return -1;
	}

	if(pDataStr = strstr(CommandsStr,"-SizeDiv(")) {
				pDataStr += 9; sscanf(pDataStr,"%lu",&SizeDiv);
				switch (SizeDiv) {
					case 1:
					case 2:
					case 4: break;
					default:
						printf("The value of \"-SizeDiv\" must be 1, 2 or 4.\n");
						return -1;
				}
	}

	if(pDataStr = strstr(CommandsStr,"-Obj(")) {
				pDataStr += 5; sscanf(pDataStr,"%lu",&ObjTotalNumber);
	} else {
		printf("The key \"-Obj\" not found.\n");
		return -1;
	}

	if(strstr(CommandsStr,"-Mask"))  { MaskMode     = true; }
	if(strstr(CommandsStr,"-SvTst")) { TestOfSaving = true; }
	if(strstr(CommandsStr,"-SvExt")) { ExtraData    = true; }

	if(strstr(CommandsStr,"-RcgMxCellBase")) {
		if(pDataStr = strstr(CommandsStr,"-RcgMxCellBase(")) {
					pDataStr += 15; sscanf(pDataStr,"%lu",&RcgMxCellBase);
		}
	}
	
	#ifndef Only_AVM_v0_5
	printf("\n================================= Ver 0.4 ===================================\n");
	MakeTest((CvAssociativeMemory32S_v0_4*)(NULL), DirStr, ObjTotalNumber, SizeDiv, MaskMode,
													TestOfSaving, ExtraData, RcgMxCellBase, 0);
	printf("\nTest is done.\n");
	#endif
	
	printf("\n================================= Ver 0.5 ===================================\n");
	MakeTest((CvAssociativeMemory32S*)(NULL), DirStr, ObjTotalNumber, SizeDiv, MaskMode,
													TestOfSaving, ExtraData, RcgMxCellBase, 1);
	printf("\nTest is done.\n");
	
	return 0;
}