//     ======================= RcgExample2.cpp ===============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 12.02.2010
//     **********************************************************************

#pragma warning (disable:4996) // This function or variable may be unsafe

#include <stdio.h>
#include "..\..\..\include\Ver0.5\AssociativeMemory.h"
#include "highgui.h"

#define cAMFileName "am_example.dat"  // File name of recognition data 
#define cObjMaxNum  4                 // Maximal number of recognition objects

int main(int argc, char* argv[])
{
	int       i;
	ULONG     FilmLengthMs  = 0;
	double    FilmFrameRate;
	IplImage* OutImg;
	CvAssociativeMemory32S* am = new CvAssociativeMemory32S;
	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== RcgExample help ===\n\n"
			   "RcgExample2.exe [file.avi | 0..9] [/r ]\n"
			   " 0..9 - capture device number\n"
			   " /r   - write result in file_out.avi\n");
		return 0;
	}

	char* argv1 = _strlwr(argv[1]);
	if(!(strstr(argv1,".avi") || isdigit(argv[1][0]))){
		printf("AVI files are processing only!\n\n");
		return -1;
	}

	#define cStrMax 256
	char FileName[cStrMax];
	char OutStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(i=2; i<argc; i++) strcat(CommandsStr,argv[i]);
	memset(FileName,0,cStrMax);
	CvCapture*     InAVI;
    IplImage*      frame = 0;
	IplImage*      gray  = 0;
	CvVideoWriter* OutAVI;

	bool WriteResultToFile = false;
	if(strstr(CommandsStr,"/r"))  WriteResultToFile = true;
	int WIDTH, HEIGHT;
	CvSize ImgSize;

    if(isdigit(argv[1][0])) {
		printf("Processing capture...\n");
        InAVI = cvCaptureFromCAM(argv[1][0] - '0');
		if( !InAVI ) {
			fprintf(stderr,"Could not initialize capturing...\n");
			return -1;
		}
		frame = cvQueryFrame(InAVI);
		ImgSize = cvGetSize(frame);
		WIDTH   = ImgSize.width;
		HEIGHT  = ImgSize.height;
		FilmFrameRate = 12;
    } else {
		printf("Processing avi file...\n");
		memcpy(FileName,argv1,strstr(argv1,".avi")-argv1);
		InAVI = cvCaptureFromFile(argv1);
		if(!InAVI) {
			printf("Can't find file %s.avi\n",FileName);
			return -1;
		}
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,1);
		FilmLengthMs  = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,0);
		WIDTH         = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_WIDTH));
		HEIGHT        = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_HEIGHT));
		FilmFrameRate = cvGetCaptureProperty(InAVI,CV_CAP_PROP_FPS);
		ImgSize       = cvSize(WIDTH, HEIGHT);
    }

	OutImg = cvCreateImage( ImgSize, 8, 3 );
	gray = cvCreateImage( ImgSize, 8, 1 );
	
	if(WriteResultToFile) {
		sprintf(OutStr,"%s_out.avi",FileName);
		OutAVI = cvCreateVideoWriter(OutStr, CV_FOURCC('c','v','i','d'), FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		if(!OutAVI) {
			OutAVI = cvCreateVideoWriter(OutStr, -1, FilmFrameRate, cvSize(WIDTH,HEIGHT), 1 );
		}
	}

	uint64 ABIndex;
	uint64 ABHitCounter;


	CvRect InterestArea;
	InterestArea.x = 84;
	InterestArea.y = 38;
	InterestArea.width  = 80;
	InterestArea.height = 80;

	if(am->Load(cAMFileName)) {
		printf("\n  Load associative bases from file: %s\n"
		   "Associative bases ......... %-7lu\n"
		   "Associative Levels ........ %-7u\n"
		   "Current base index ........ %-7I64u\n"
		   "Wr/Rd counter ............. %-7I64u\n\n",
		   cAMFileName, am->GetTotalABases(),
		   am->GetTotalLevels(), am->GetCurIndex(), am->GetWrRdCounter());
	} else { am->Create(cvSize(InterestArea.width, InterestArea.height)); }


	cvNamedWindow("Result", 1);

	long PauseLoop    = 200;
	long TrainingLoop = 600;

	for(;;) {
		
		frame = cvQueryFrame(InAVI);
		if( !frame ) break;

		cvFlip(frame);

		cvCopy(frame, OutImg);
		
		cvCvtColor(frame, gray, CV_BGR2GRAY);

		if(PauseLoop) {
			cvRectangle( OutImg, cvPoint(InterestArea.x,   InterestArea.y),
							 cvPoint(InterestArea.x + InterestArea.width,
							 InterestArea.y + InterestArea.height), cvScalar(0,255,0),2,8,0);
			PauseLoop--;
		} else {
			am->SetImage(gray);
			if(TrainingLoop) {
				// Training
				long ObjIndex = cObjMaxNum;// <= Actual index of training object (Now it lastest object)
				CvRect SAr;
				am->Write(InterestArea, &ObjIndex);
				if(am->Read(InterestArea, &SAr, 0, &ABIndex, &ABHitCounter)) {
					printf("Hitting => Associative base index = %I64u, Hit counter = %I64u, Total bases = %lu\n",
					   ABIndex, ABHitCounter, am->GetTotalABases());					
				}
				cvRectangle( OutImg, cvPoint(InterestArea.x,   InterestArea.y),
								 cvPoint(InterestArea.x + InterestArea.width,
								 InterestArea.y + InterestArea.height), cvScalar(0,0,255),2,8,0);
				TrainingLoop--;
			} else {
				// Object recognition
				CvMemStorage* pMemStorage = cvCreateMemStorage();

				//CvSeq* pSeq = am->ObjectRecognition(pMemStorage);
				CvSeq* pSeq = am->ObjectTracking(pMemStorage);

				char OutNumStr[256];
				CvFont font;
				cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );
				static char ObjName[cObjMaxNum][64] = { "Mr. Dmitriy", "Toy \"Heart\"",
					                                    "Book", "Post card"/* "your name of object", ... */};
				CvObjDsr32S ObjDsr;
				CvSeqReader SegReader;
				cvStartReadSeq(pSeq, &SegReader, 0);

				for(int i=0; i<pSeq->total; i++) {
					CV_READ_SEQ_ELEM(ObjDsr, SegReader);

					cvRectangle(OutImg, cvPoint(ObjDsr.ObjRect.x,   ObjDsr.ObjRect.y),
						cvPoint(ObjDsr.ObjRect.x + ObjDsr.ObjRect.width,
						ObjDsr.ObjRect.y + ObjDsr.ObjRect.height),
						cvScalar(255,0,0),2,8,0);
					
					sprintf(OutNumStr,"%s",ObjName[ObjDsr.Data-1]);
					cvPutText(OutImg, OutNumStr, cvPoint(ObjDsr.ObjRect.x+10, ObjDsr.ObjRect.y +
				                               ObjDsr.ObjRect.height+24), &font, CV_RGB(255,255,0));

					if(ObjDsr.Trj.Age > 2) {
						for(int t=1; t<ObjDsr.Trj.Age; t++) {
							cvLine(OutImg, ObjDsr.Trj.Pnt[cAM_TrcLen-t-1],
								ObjDsr.Trj.Pnt[cAM_TrcLen-t], cvScalar(255,255,0),2,8,0);
						}
						cvCircle(OutImg, ObjDsr.Trj.Pnt[cAM_TrcLen-1],4,cvScalar(0,255,0),2,8,0);
					}
				}

				cvReleaseMemStorage( &pMemStorage );
			}
		}

		ULONG POS_MSEC = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
		int hour        = int(POS_MSEC/(3600000));
		POS_MSEC        %= 3600000L;
		int minute      = int(POS_MSEC/(60000));
		POS_MSEC        %= 60000L;
		int second      = int(POS_MSEC/1000);
		int millisecond = int(POS_MSEC % 1000);
		printf("==> film current position: %2.2d:%2.2d:%2.2d.%3.3d\n",
				hour, minute, second, millisecond);

	    cvShowImage("Result", OutImg);

		if(WriteResultToFile) { cvWriteFrame(OutAVI, OutImg); }

		if( cvWaitKey(10) >= 0 ) break;
	}

	// am->Save(cAMFileName);

	delete am;

	if(WriteResultToFile) { cvReleaseVideoWriter(&OutAVI); }
	
	cvReleaseCapture(&InAVI);
	cvReleaseImage(&gray);
	
	return 0;
}
