//     ========================= SimilarityCmp.cpp ==========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 03.02.2010
//     **********************************************************************

#include "..\..\..\include\Ver0.4\AssociativeMemory.h"
#include "..\..\..\include\AVM_drawing.h"
#include "cv.h"
#include "cxmisc.h"

#define cDiagramPointSize 4
#define cBorderSize       70
#define cDiagramYSize     300
#define cPassMaxNum       2

double Similarity = 0;

void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args)
{
	static char ObjName[10][64] = { "Postcard", "Hand-close",
		"Book", "Post card", "Body spray", "Toy \"Dog\"", "Toy \"Pig\"",
		"Clock", "Hand", "Test area"};
	char OutNumStr[256];

	int i;
	static double bTime,eTime,Tf;
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );

	IplImage* pImg = (IplImage*)args;

	switch(aState) {
		case cProcessingIsBegun:
			bTime = (double)cvGetTickCount();
			Tf = cvGetTickFrequency()*1000.0;
			break;

		
		case cRecognizedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x, apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,128,0),1,8,0);
			printf("R   Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);
			break;

		case cGeneralizedObject:
		case cTrackedObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x + apParam->ObjRect.width,
									   apParam->ObjRect.y + apParam->ObjRect.height),
									   cvScalar(255,0,0),2,8,0);

			sprintf(OutNumStr,"%s",ObjName[*apData-1]);
			cvPutText(pImg, OutNumStr, cvPoint(apParam->ObjRect.x+10, apParam->ObjRect.y +
				                               apParam->ObjRect.height+24), &font, CV_RGB(255,255,0));

			printf("G-T Obj: [%s] ...... Similarity: %7.4f\n", ObjName[*apData-1], apParam->Similarity);

			if(apParam->ppUserTrackingInfo) {
				#define cTrcLen 100
				struct TTrcInfo {
					CvPoint TrcPoint[cTrcLen];
					long Age;
				};
				if(!*apParam->ppUserTrackingInfo) {
					*apParam->ppUserTrackingInfo = new TTrcInfo;
					memset(*apParam->ppUserTrackingInfo, 0, sizeof(TTrcInfo));
				} else {
					TTrcInfo* pTrcInfo = (TTrcInfo*)*apParam->ppUserTrackingInfo;
					memcpy(&pTrcInfo->TrcPoint[0], &pTrcInfo->TrcPoint[1], sizeof(CvPoint)*(cTrcLen-1));
					pTrcInfo->TrcPoint[cTrcLen-1] = cvPoint(apParam->ObjRect.x + (apParam->ObjRect.width>>1),
						                                    apParam->ObjRect.y + (apParam->ObjRect.height>>1));
					if(pTrcInfo->Age < cTrcLen) { pTrcInfo->Age++; }

					if(pTrcInfo->Age > 2) {
						for(i=1; i<pTrcInfo->Age; i++) {
							cvLine(pImg, pTrcInfo->TrcPoint[cTrcLen-i-1],
								         pTrcInfo->TrcPoint[cTrcLen-i], cvScalar(255,255,0),2,8,0);
						}
						cvCircle(pImg, pTrcInfo->TrcPoint[cTrcLen-1],4,cvScalar(0,255,0),2,8,0);
					}
				}
			}
			Similarity = (Similarity + apParam->Similarity)/2.0;
			break;

		case cDeletingTrackedObject:
			if(*apParam->ppUserTrackingInfo) { delete *apParam->ppUserTrackingInfo; }
			break;

		case cLearnThisObject:
			cvRectangle( pImg, cvPoint(apParam->ObjRect.x,   apParam->ObjRect.y),
							   cvPoint(apParam->ObjRect.x +  apParam->ObjRect.width,
								       apParam->ObjRect.y +  apParam->ObjRect.height),
									   cvScalar(255,255,0),2,8,0);
			printf("L   Obj: [%s] ...... Similarity: %7.4f, Bases: %lu\n",
				   ObjName[*apData-1], apParam->Similarity, apParam->TotalABases);
			break;

		case cProcessingIsFinished:
			eTime = (double)cvGetTickCount();
			sprintf(OutNumStr,"%5.1f ms",(eTime-bTime)/Tf);
			cvPutText(pImg, OutNumStr, cvPoint(0, 24), &font, CV_RGB(0,255,0));
			break;
	}
}

bool doShowABases(CvSize  aRcgMxSize, short* apRcgMx, long* apData,
				  uint64 aIndex, uint64 aHitCounter,
				  int aContainBasesTotal, short  aLevel, void* args)
{
	IplImage* pImg = (IplImage*)args;
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );
	char OutNumStr[256];
	
	int sX = pImg->width/aRcgMxSize.width;
	int sY = pImg->height/aRcgMxSize.height;
	
	for(int y=0,iY=0; y<aRcgMxSize.height; y++,iY+=sY) {
		for(int x=0,iX=0; x<aRcgMxSize.width; x++,iX+=sX) {
			short d = apRcgMx[y*aRcgMxSize.width+x];
			uchar hC = (d + 32768)?uchar(128 + (127*int(d))/32768):255;
			cvRectangle(pImg, cvPoint(iX, iY), cvPoint(iX + sX, iY + sY), cvScalar(hC,hC,hC),CV_FILLED,8,0);
		}
	}
	sprintf(OutNumStr,"Level %u",aLevel);
	cvPutText(pImg, OutNumStr, cvPoint(10, 30), &font, CV_RGB(0,255,0));
	sprintf(OutNumStr,"Size %ux%u", aRcgMxSize.width, aRcgMxSize.height);
	cvPutText(pImg, OutNumStr, cvPoint(10, 60), &font, CV_RGB(0,255,0));
	sprintf(OutNumStr,"Index %I64u", aIndex);
	cvPutText(pImg, OutNumStr, cvPoint(10, 90), &font, CV_RGB(0,255,0));
	sprintf(OutNumStr,"Bases %d", aContainBasesTotal);
	cvPutText(pImg, OutNumStr, cvPoint(10, 120), &font, CV_RGB(0,255,0));
	sprintf(OutNumStr,"Hits %I64u", aHitCounter);
	cvPutText(pImg, OutNumStr, cvPoint(10, 150), &font, CV_RGB(0,255,0));
	sprintf(OutNumStr,"Data %d", *apData);
	cvPutText(pImg, OutNumStr, cvPoint(10, 180), &font, CV_RGB(0,255,0));
	
	cvShowImage("Result", pImg);
	cvWaitKey(200);

	return true;
}

char* TimeToStr(long aPOS_MSEC)
{
	static char TimeStr[64];
	int hour    = int(aPOS_MSEC/(3600000));
	aPOS_MSEC  %= 3600000L;
	int minute  = int(aPOS_MSEC/(60000));
	aPOS_MSEC  %= 60000L;
	int second  = int(aPOS_MSEC/1000);
	int millisecond = int(aPOS_MSEC % 1000);
	sprintf(TimeStr, "%2.2d:%2.2d:%2.2d.%3.3d", hour, minute, second, millisecond);
	return TimeStr;
}


#define _am(Method) (Pass)?am->Method:am_v0_4->Method


int main(int argc, char* argv[])
{
	int       i;
	ULONG     FilmLengthMs  = 0;
	double    FilmFrameRate;
	IplImage* OutImg;
	IplImage* GrayImg;
	static CvScalar sColorForPass[cPassMaxNum] = { {0,0,255}, {0,255,255} };

	CvAssociativeMemory32S* am = new CvAssociativeMemory32S;
	CvAssociativeMemory32S_v0_4* am_v0_4 = new CvAssociativeMemory32S_v0_4;

	
	if(argc <= 1 || (argc == 2 && strstr(argv[1],"/?"))) {
		printf("\n\n=== Similarity comparing help ===\n\n"
			   "SimilarityCmp.exe [path]RcgTraining4s.avi\n");
		return 0;
	}

	char* argv1 = _strlwr(argv[1]);
	if(!(strstr(argv1,".avi") || isdigit(argv[1][0]))){
		printf("AVI files are processing only!\n\n");
		return -1;
	}

	#define cStrMax 256
	char FileName[cStrMax];
	char OutStr[cStrMax];
	char CommandsStr[cStrMax];
	memset(CommandsStr,0,cStrMax);
	for(i=2; i<argc; i++) strcat(CommandsStr,argv[i]);
	memset(FileName,0,cStrMax);
	CvCapture*     InAVI;
    IplImage*      frame = 0;

	int WIDTH, HEIGHT;
	CvSize ImgSize;

	if(!strstr(argv1,"rcgtraining4s.avi")) {
		printf("For testing the file \"RcgTraining4s\".avi needed.\n");
		return -1;
	}
	
	printf("Processing avi file...\n");
	memcpy(FileName,argv1,strstr(argv1,".avi")-argv1);
	InAVI = cvCaptureFromFile(argv1);

	if(!InAVI) {
		printf("Can't find file %s.avi\n",FileName);
		return -1;
	}

	cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,1);
	FilmLengthMs  = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
	cvSetCaptureProperty(InAVI,CV_CAP_PROP_POS_AVI_RATIO,0);
	WIDTH         = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_WIDTH));
	HEIGHT        = int(cvGetCaptureProperty(InAVI,CV_CAP_PROP_FRAME_HEIGHT));
	FilmFrameRate = cvGetCaptureProperty(InAVI,CV_CAP_PROP_FPS);
	ImgSize       = cvSize(WIDTH, HEIGHT);

	OutImg  = cvCreateImage( ImgSize, 8, 3 );
	GrayImg = cvCreateImage( ImgSize, 8, 1 );


	int Pass = 0;
	int TotalPoints = 100;
	int PointInterval = int(ceil((FilmLengthMs*FilmFrameRate)/(1000*TotalPoints)));
	int ResImgW = TotalPoints*cDiagramPointSize + cBorderSize*2;
	int ResImgH = cDiagramYSize + cBorderSize*2;
	CvFont BigFont;
	cvInitFont( &BigFont, 2, 0.8, 0.8, 0, 2, 8 );
	CvFont MiddleFont;
	cvInitFont( &MiddleFont, 2, 0.7, 0.7, 0.3, 1, 8 );
	CvFont SmallFont;
	cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );

	sprintf(OutStr,"Similarity_Test_%u_ms_%ux%u_pass1.png", FilmLengthMs, WIDTH, HEIGHT);

	IplImage* ResImage = cvLoadImage(OutStr);

	int DiagramXBegin    = cBorderSize;
	int DiagramXEnd      = ResImgW-cBorderSize;
	int DiagramYTop      = ResImgH-(cDiagramYSize + cBorderSize);
	int DiagramYMiddle   = ResImgH-(cDiagramYSize/2 + cBorderSize);
	int DiagramYBottom   = ResImgH-cBorderSize;
	int DiagramYBottomT1 = DiagramYBottom + 16;
	int DiagramYBottomT2 = DiagramYBottom + 40;

	if(!ResImage) {
		ResImage = cvCreateImage(cvSize(ResImgW, ResImgH), IPL_DEPTH_8U, 3);
		cvSet(ResImage,cvScalarAll(180),0);
		
		cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYBottom),
						cvPoint(DiagramXEnd,   DiagramYBottom), CV_RGB(0,0,0),1,8,0);
		cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYMiddle),
						cvPoint(DiagramXEnd,   DiagramYMiddle), CV_RGB(0,0,0),1,8,0);
		cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYTop),
						cvPoint(DiagramXEnd,   DiagramYTop),    CV_RGB(0,0,0),1,8,0);
		cvLine(ResImage,cvPoint(DiagramXBegin, DiagramYTop),
						cvPoint(DiagramXBegin, DiagramYBottom), CV_RGB(0,0,0),1,8,0);
		cvLine(ResImage,cvPoint(DiagramXEnd,   DiagramYTop),
						cvPoint(DiagramXEnd,   DiagramYBottom), CV_RGB(0,0,0),1,8,0);

		cvPutText(ResImage, "Similarity diagram", cvPoint(160, DiagramYTop-20), &MiddleFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "0.0", cvPoint(2, DiagramYBottom), &BigFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "0.5", cvPoint(2, DiagramYMiddle), &BigFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "1.0", cvPoint(2, DiagramYTop),    &BigFont, CV_RGB(0,0,0));
		sprintf(OutStr, "%35s.avi   %s", FileName, TimeToStr(FilmLengthMs));
		cvPutText(ResImage, OutStr, cvPoint(DiagramXBegin, DiagramYBottomT1),&SmallFont, CV_RGB(0,0,0));
		cvPutText(ResImage, "Ver0.4", cvPoint(DiagramXBegin, DiagramYBottomT2),&SmallFont, sColorForPass[0]);
		cvPutText(ResImage, "Ver0.5", cvPoint(DiagramXBegin+70, DiagramYBottomT2),&SmallFont, sColorForPass[1]);
	} else {
		Pass = 1;	
	}
	
	cvNamedWindow("Diagram", 1);
	cvNamedWindow("Result", 1);

	CvDrawAssociativeTree DrawTree(am);
	
	uint64 ABIndex;
	uint64 ABHitCounter;
	double BestSimilarity = 0;

	CvRect InterestArea;
	InterestArea.x = 84;
	InterestArea.y = 38;
	InterestArea.width  = 80;
	InterestArea.height = 80;

	am_v0_4->Create(cvSize(InterestArea.width, InterestArea.height));

	double RcgMxExpRate_A = 0.07;
	double RcgMxExpRate_B = 0.2;
	double RcgMxExpRate_C = 0.015;
	int MaxClusterAge = 40;

//	for(double RcgMxExpRate_A = 0.05; RcgMxExpRate_A < 0.09; RcgMxExpRate_A += 0.01) {
//		for(double RcgMxExpRate_B = 0.1; RcgMxExpRate_B < 0.5; RcgMxExpRate_B += 0.1) {
//			for(double RcgMxExpRate_C = 0.005; RcgMxExpRate_C < 0.007; RcgMxExpRate_C += 0.02) {
				for(int r=0; r<1; r++) {
					static int ClstAge[2] = { 40, 128 };
					MaxClusterAge = ClstAge[r];

					if(Pass) {
						sprintf(OutStr,"Similarity_Test_%u_ms_%ux%u_pass1.png", FilmLengthMs, WIDTH, HEIGHT);
						if(ResImage) { cvReleaseImage(&ResImage); }
						ResImage = cvLoadImage(OutStr);
					}

					int IntervalCounter = 0;
					int PointNum = 0;
					CvPoint sPrevPoint, sNextPoint;
					long TrainingLoop = 400;
					Similarity = 0;
					double SimilaritySum = 0;
					
					am->Create(cvSize(InterestArea.width, InterestArea.height));
					
					am->SetParam(ptRcgMxCellBase, 21800);
					am->SetParam(ptRcgMxExpRate_A, RcgMxExpRate_A);
					am->SetParam(ptRcgMxExpRate_B, RcgMxExpRate_B);
					am->SetParam(ptRcgMxExpRate_C, RcgMxExpRate_C);
					am->SetParam(ptMaxClusterAge, MaxClusterAge);

					for(;;) {
						if(!cvGrabFrame(InAVI)) {
							int PrevPass = Pass;

							if(!Pass) {
								Pass++;
								sprintf(FileName,"Similarity_Test_%u_ms_%ux%u_pass1.png", FilmLengthMs, WIDTH, HEIGHT);
							} else {
								double ActSimilarity = SimilaritySum/PointNum;
								sprintf(OutStr, "%7.5f", ActSimilarity);
								cvPutText(ResImage, OutStr, cvPoint(DiagramXEnd-70, DiagramYTop-10), &BigFont, CV_RGB(0,0,0));
								sprintf(OutStr, "Total bases: %-3lu | %u %4.2f %4.2f %5.3f",
										am->GetTotalABases(), MaxClusterAge, RcgMxExpRate_A, RcgMxExpRate_B, RcgMxExpRate_C);
								cvPutText(ResImage, OutStr, cvPoint(DiagramXBegin+130, DiagramYBottomT2),&SmallFont, CV_RGB(0,0,0));
								sprintf(FileName,"Similarity_Test_%u_%4.2f_%4.2f_%4.2f_%u_ms_%ux%u.png",
									MaxClusterAge, RcgMxExpRate_A, RcgMxExpRate_B, RcgMxExpRate_C, FilmLengthMs, WIDTH, HEIGHT);
								if(BestSimilarity < ActSimilarity) {
									BestSimilarity = ActSimilarity;
									cvNamedWindow("Best result", 1);
									cvShowImage("Best result", ResImage);
								}
							}
							if(cvSaveImage(FileName, ResImage)) {
								printf("\nResult image was saved into file %s\n\n", FileName);
							}
							cvSetCaptureProperty( InAVI, CV_CAP_PROP_POS_MSEC, 0 );
							if(PrevPass) { break; } else {
								IntervalCounter = 0;
								PointNum        = 0;
								TrainingLoop    = 400;
								Similarity      = 0;
								SimilaritySum   = 0;
								continue;
							}
						} 
						
						frame = cvRetrieveFrame(InAVI);

						cvFlip(frame);

						cvCopy(frame, OutImg);
						
						cvCvtColor(frame, GrayImg, CV_BGR2GRAY);

						_am(SetImage(GrayImg));

						if(TrainingLoop) {

							//double EOT = _am(EstimateOpportunityForTraining(InterestArea));
							//printf("EOT = %f\n", EOT);
							
							long Data = 1;
							CvRect SAr;
							_am(Write(InterestArea, &Data));
							double ActSimilarity;
							_am(Read(InterestArea, &SAr, 0, &ABIndex, &ABHitCounter, &ActSimilarity, true)); 
							printf("Hitting! Index = %I64u, HitCounter = %I64u, TotalABases = %lu, Similarity = %f\n",
							   ABIndex, ABHitCounter, _am(GetTotalABases()), ActSimilarity);
							cvRectangle( OutImg, cvPoint(InterestArea.x,   InterestArea.y),
											 cvPoint(InterestArea.x + InterestArea.width,
											 InterestArea.y + InterestArea.height), cvScalar(0,0,255),2,8,0);
							Similarity = (Similarity + ActSimilarity)/2.0;
							TrainingLoop--;
						} else {
							//break;//!!!
							_am(ObjectTracking(doSearchObject, OutImg));
						}

						static short DivCounter;
						if(!(DivCounter++ % 12)) {
							DrawTree.ShowAssociativeTree();
						}

						ULONG POS_MSEC = ULONG(cvGetCaptureProperty(InAVI, CV_CAP_PROP_POS_MSEC));
						int hour        = int(POS_MSEC/(3600000));
						POS_MSEC        %= 3600000L;
						int minute      = int(POS_MSEC/(60000));
						POS_MSEC        %= 60000L;
						int second      = int(POS_MSEC/1000);
						int millisecond = int(POS_MSEC % 1000);
						printf("==> film current position: %2.2d:%2.2d:%2.2d.%3.3d\n",
								hour, minute, second, millisecond);

						cvShowImage("Result", OutImg);

						IntervalCounter = (++IntervalCounter) % PointInterval;

						if(!IntervalCounter) {
							sNextPoint = cvPoint(PointNum*cDiagramPointSize + cBorderSize,
												 ResImgH - cBorderSize - int(cDiagramYSize*Similarity));
							if(PointNum) {
								cvLine(ResImage, sPrevPoint, sNextPoint, sColorForPass[Pass], cDiagramPointSize,8,0);
							}
							SimilaritySum += Similarity;
							sPrevPoint = sNextPoint;
							PointNum++;
						}
						cvShowImage("Diagram", ResImage);

						if( cvWaitKey(10) >= 0 ) break;
					}
				}
//			}
//		}
//	}
	


	am->OptimizeAssociativeTree();
	DrawTree.ShowAssociativeTree();

	cvWaitKey();

//	am->DoForEach(doShowABases, OutImg);

	delete am;
	delete am_v0_4;

	cvReleaseCapture(&InAVI);

	cvReleaseImage(&GrayImg);
	cvReleaseImage(&OutImg);
	
	return 0;
}
